@@ -1,32 +1,19 @@
-/*	$Id: engine.c,v 1.27 2006/02/07 10:06:31 mbalmer Exp $	*/
+/*	$Id: engine.c,v 1.37 2006/02/08 14:47:28 mbalmer Exp $	*/
 
 /*
- * Copyright (c) 2003-2006 Marc Balmer <marc@msys.ch>.
- * All rights reserved.
+ * Copyright (c) 2003, 2004, 2005, 2006 Marc Balmer <marc@msys.ch>.
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
+ * Permission to use, copy, modify, and distribute this software for any
+ * purpose with or without fee is hereby granted, provided that the above
+ * copyright notice and this permission notice appear in all copies.
  *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above
- *    copyright notice, this list of conditions and the following
- *    disclaimer in the documentation and/or other materials provided
- *    with the distribution.
- * 3. The name of the author may not be used to endorse or promote products
- *    derived from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
- * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
- * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
- * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
- * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
- * POSSIBILITY OF SUCH DAMAGE.
+ * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+ * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+ * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+ * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+ * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
 #include <sys/time.h>
@@ -119,14 +106,21 @@ extern FILE	*logfp;
 extern char	*user;
 extern char	*group;
 extern char	*chrootdir;
-extern int	 rlimit_nofile;
+extern rlim_t	 rlimit_nofile;
+extern rlim_t	 rlimit_nproc;
 extern int	 nbe;
 extern char	*spamprefix;
 extern int	 logvirus, logspam, logunwanted;
 extern int	 keep_tmpfiles;
+extern int	 ex_timeout;
 
 struct imsgbuf	*ibuf_main;
 
+/* For ceonnection reference counting */
+
+static int		conn_cnt = 0;
+static pthread_mutex_t	conn_cnt_mutex = PTHREAD_MUTEX_INITIALIZER;
+
 extern char *__progname;
 
 #ifdef ENABLE_LDAP
@@ -207,6 +201,15 @@ static void pf_table_clr(const char *, const char *, long);
 static void imsg_commit();
 #endif
 
+void
+e_sighdlr(int signum)
+{
+	switch (signum) {
+	case SIGHUP:
+		break;
+	}
+}
+
 #ifdef ENABLE_LDAP
 /*
  * Decode strategies represented as string values
@@ -281,6 +284,15 @@ connection_setup(SMFICTX *ctx)
 		return SMFIS_TEMPFAIL;
 	}
 
+	if (pthread_mutex_lock(&conn_cnt_mutex))
+		syslog(LOG_ERR, "unable to increase connection count");
+	else {
+		++conn_cnt;
+		if (verbose)
+			warnx("connection increased to %d", conn_cnt);
+		pthread_mutex_unlock(&conn_cnt_mutex);
+	}
+
 	return SMFIS_CONTINUE;
 }
 
@@ -403,6 +415,17 @@ connection_cleanup(SMFICTX *ctx)
 		smfi_setpriv(ctx, NULL);
 	} else
 		syslog(LOG_ERR, "unable to get internal data in connection_cleanup()");
+
+	if (pthread_mutex_lock(&conn_cnt_mutex))
+		syslog(LOG_ERR, "can't decrease connection count");
+	else {
+		--conn_cnt;
+		if (verbose)
+			warnx("connection count decreased to %d", conn_cnt);
+		if (conn_cnt < 0)
+			syslog(LOG_ERR, "connection count drops below zero");
+		pthread_mutex_unlock(&conn_cnt_mutex);
+	}
 }
 
 static sfsistat
@@ -1592,6 +1615,8 @@ e_main(uid_t uid, gid_t gid, int pipe_m2e[2])
 	struct rlimit	 rl;
 	int		 n;
 	int		 pid;
+	int		 nconn;
+	int		 tout;
 
 	switch (pid = fork()) {
 	case -1:
@@ -1612,7 +1637,14 @@ e_main(uid_t uid, gid_t gid, int pipe_m2e[2])
 		rl.rlim_cur = rl.rlim_max = rlimit_nofile;
 
 		if (setrlimit(RLIMIT_NOFILE, &rl))
-			err(1, "can set number of files to %d", rlimit_nofile);
+			err(1, "can set number of files to %lld", (long long)rlimit_nofile);
+	}
+
+	if (rlimit_nproc > 0) {
+		rl.rlim_cur = rl.rlim_max = rlimit_nproc;
+
+		if (setrlimit(RLIMIT_NPROC, &rl))
+			err(1, "can set number of processes to %lld", (long long)rlimit_nproc);
 	}
 
 	if (logfile != NULL && (logfp = fopen(logfile, "a")) == NULL)
@@ -1620,15 +1652,11 @@ e_main(uid_t uid, gid_t gid, int pipe_m2e[2])
 	else
 		logfp = NULL;
 
-	if (verbose)
-		warnx("not writing a pid file when running in verbose mode");
-	else {
-		if ((fp = fopen(pidfile, "w")) == NULL) 
-			err(1, "can't open pidfile %s", pidfile);
-		if (fprintf(fp, "%d\n", getpid()) == -1)
-			err(1, "can't write pid to pidfile %s", pidfile);
-		fclose(fp);
-	}
+	if ((fp = fopen(pidfile, "w")) == NULL) 
+		err(1, "can't open pidfile %s", pidfile);
+	if (fprintf(fp, "%d\n", getpid()) == -1)
+		err(1, "can't write pid to pidfile %s", pidfile);
+	fclose(fp);
 
 	/* chroot, if necessary */
 
@@ -1688,6 +1716,37 @@ e_main(uid_t uid, gid_t gid, int pipe_m2e[2])
 	if (smfi_main() == MI_FAILURE)
 		syslog(LOG_ERR, "failed to establish a connection on port %s", port);
 
+	/* Wait for the connection reference count to drop to zero */
+
+	if (verbose)
+		warnx("gracefully terminating");
+
+	if (ex_timeout > 0)
+		tout = ex_timeout;
+
+	nconn = 0;
+	do {
+		if (pthread_mutex_lock(&conn_cnt_mutex))
+			syslog(LOG_ERR, "can get connection count");
+		else {
+			nconn = conn_cnt;
+			pthread_mutex_unlock(&conn_cnt_mutex);
+		}
+
+		if (nconn > 0) {
+			if (verbose)
+				warnx("waiting for connection count to reach zero (now %d)", nconn);
+			sleep(1);
+			if (ex_timeout > 0) {
+				if (tout-- == 0)
+					nconn = 0;
+			}
+		}
+	} while (nconn > 0);
+
+	if (verbose)
+		warnx("connection count reached zero, terminating instance");
+
 	for (n = 0; n < nbe; n++) {
 		if (be[n].be_exit != NULL)
 			be[n].be_exit();
@@ -1696,7 +1755,7 @@ e_main(uid_t uid, gid_t gid, int pipe_m2e[2])
 	}
 	free(be);
 
-	syslog(LOG_INFO, "exiting");
+	syslog(LOG_INFO, "child process exiting");
 
 	closelog();
 	if (logfp != NULL)
