@@ -1,32 +1,19 @@
-/*	$Id: smtp-vilter.c,v 1.131 2006/02/06 10:33:49 mbalmer Exp $	*/
+/*	$Id: smtp-vilter.c,v 1.143 2006/02/09 11:17:39 mbalmer Exp $	*/
 
 /*
- * Copyright (c) 2003-2006 Marc Balmer <marc@msys.ch>.
- * All rights reserved.
+ * Copyright (c) 2003, 2004, 2005, 2006 Marc Balmer <marc@msys.ch>
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
+ * Permission to use, copy, modify, and distribute this software for any
+ * purpose with or without fee is hereby granted, provided that the above
+ * copyright notice and this permission notice appear in all copies.
  *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above
- *    copyright notice, this list of conditions and the following
- *    disclaimer in the documentation and/or other materials provided
- *    with the distribution.
- * 3. The name of the author may not be used to endorse or promote products
- *    derived from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
- * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
- * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
- * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
- * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
- * POSSIBILITY OF SUCH DAMAGE.
+ * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+ * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+ * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+ * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+ * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
 #include <sys/time.h>
@@ -122,7 +109,8 @@ FILE	*logfp;
 char	*user;
 char	*group;
 char	*chrootdir;
-int	 rlimit_nofile;
+rlim_t	 rlimit_nofile;
+rlim_t	 rlimit_nproc;
 int	 nbe;
 char	*spamprefix;
 int	 logvirus, logspam, logunwanted;
@@ -146,7 +134,11 @@ char	*binddn;
 char	*bindpasswd;
 int	 ldap_use_tls;
 
+int	 ex_timeout;	/* Exit timeout waiting for pending connections */
+
 volatile sig_atomic_t quit = 0;
+volatile sig_atomic_t child_quit = 0;
+volatile sig_atomic_t reconfig = 0;
 
 struct backend *be;
 int	 dev = -1;
@@ -166,11 +158,17 @@ static void
 sighdlr(int signum)
 {
 	switch (signum) {
+	case SIGCHLD:
+		child_quit = 1;
+		break;
 	case SIGINT:
 	case SIGTERM:
-	case SIGCHLD:
 		quit = 1;
 		break;
+	case SIGHUP:
+		/* reconfigure */
+		reconfig = 1;
+		break;
 	}
 }
 
@@ -181,16 +179,42 @@ static void
 #endif
 usage(void)
 {
-	fprintf(stderr, "usage: %s [-v] [-V] [-m] [-k] [-p port] [-d dir] "
-	    "[-T tmpfile-opt] [-e logile] [-b backend] "
-	    "[-n recipient-notification] [-o notify-only] [-C configfile] "
-	    "[-P pidfile] [-t chroot-dir] [-u user] [-g group] [-f maxfiles] "
-	    "[-a spam-subject-prefix]", __progname);
+	fprintf(stderr, "usage: %s"
+	    " [-Vvmkx]"
+	    " [-A path]"
+	    " [-a spam-subject-prefix]"
 #ifdef ENABLE_LDAP
-	fprintf(stderr, " [-D binddn] [-h ldaphost] [-L ldapport] "
-	    "[-U ldapurl] [-w bindpasswd] [-B searchbase]");
+	    " [-B searchbase]"
 #endif
-	fprintf(stderr, "\n");
+	    " [-b backend]"
+	    " [-C configfile]"
+	    " [-d dir] "
+#ifdef ENABLE_LDAP
+	    " [-D binddn]"
+#endif
+	    " [-E timeout]"
+	    " [-e logile]"
+	    " [-f maxfiles]"
+	    " [-g group]"
+#ifdef ENABLE_LDAP
+	    " [-h ldaphost]"
+	    " [-L ldapport]"
+#endif
+	    " [-n recipient-notification]"
+	    " [-o notify-only]" 
+	    " [-P pidfile]"
+ 	    " [-p port]"
+	    " [-T tmpfile-opt]"
+	    " [-t chroot-dir]"
+#ifdef ENABLE_LDAP
+	    " [-U ldapurl]"
+#endif
+	    " [-u user]"
+#ifdef ENABLE_LDAP
+	    " [-w bindpasswd]"
+#endif
+	    "\n"
+		, __progname);
 	exit(1);
 }
 
@@ -252,6 +276,7 @@ main(int argc, char *argv[])
 	struct pollfd	 pfd[1];
 	int		 n;
 	pid_t		 child_pid, pid;
+	char		*fnam;	/* file to analyze */
 #ifdef ENABLE_LDAP
 	int		 version;
 #endif
@@ -279,6 +304,8 @@ main(int argc, char *argv[])
 	unwanted_strategy = -1;
 	error_strategy = -1;
 
+	fnam = NULL;
+
 	virus_reaction = REACTION_UNSET;
 	spam_reaction = REACTION_UNSET;
 	unwanted_reaction = REACTION_UNSET;
@@ -292,10 +319,12 @@ main(int argc, char *argv[])
 	recipient_notification = NULL;
 	notify_only = NULL;
 	markall = 0;
-	rlimit_nofile = -1;
+	rlimit_nofile = 0;
+	rlimit_nproc = 0;
 	spamprefix = NULL;
 	logvirus = logspam = logunwanted = 0;
 	keep_tmpfiles = 0;
+	ex_timeout = -1;
 #ifdef ENABLE_LDAP
 	ldaphost = NULL;
 	ldapport = -1;
@@ -310,23 +339,88 @@ main(int argc, char *argv[])
 	/* Process the commandline */
 
 #ifdef ENABLE_LDAP	
-	while ((ch = getopt(argc, argv, "vmp:d:T:b:e:n:o:C:P:t:u:U:g:Vf:ka:xD:h:L:w:B:?")) != -1) {
+	while ((ch = getopt(argc, argv, "A:a:B:b:C:D:d:E:e:f:g:h:kL:mn:o:P:p:T:t:u:U:Vvw:x?")) != -1) {
 #else
-	while ((ch = getopt(argc, argv, "vmp:d:T:b:e:n:o:C:P:t:u:g:Vf:ka:x?")) != -1) {
+	while ((ch = getopt(argc, argv, "A:a:b:C:d:E:e:f:g:kmn:o:P:p:T:t:u:Vvx?")) != -1) {
 #endif
 		switch (ch) {
-		case 'v':
-			++verbose;
+		case 'A':
+			fnam = optarg;
 			break;
+		case 'a':
+			spamprefix = optarg;
+			break;
+#ifdef ENABLE_LDAP
+		case 'B':
+			searchbase = optarg;
+			break;
+#endif
+		case 'b':
+			backend = optarg;
+			decode_backend(backend);
+			break;
+		case 'C':
+			cfgfile = optarg;
+			break;
+#ifdef ENABLE_LDAP
+		case 'D':
+			binddn = optarg;
+			break;
+#endif
+		case 'd':
+			tmpdir = optarg;
+			break;
+		case 'E':	
+			ex_timeout = (int)strtonum(optarg, 1LL, MAX_TIMEOUT, &errstr);
+			if (errstr)
+				errx(1, "timeout is %s: %s", errstr, optarg);
+			break;
+		case 'e':
+			logfile = optarg;
+			break;
+		case 'f':
+			if (getuid()) {
+				getrlimit(RLIMIT_NOFILE, &rl);
+				rlimit_max = rl.rlim_max;
+			} else
+				rlimit_max = RLIM_INFINITY;
+			rlimit_nofile = (int)strtonum(optarg, 1LL, rlimit_max, &errstr);
+			if (errstr)
+				errx(1, "number of files is %s: %s", errstr, optarg);
+			break; 
+		case 'g':
+			group = optarg;
+			break;
+#ifdef ENABLE_LDAP
+		case 'h':
+			ldaphost = optarg;
+			break;
+#endif
+		case 'k':
+			keep_tmpfiles = 1;
+			break;
+#ifdef ENABLE_LDAP
+		case 'L':
+			ldapport = strtonum(optarg, 1, 65535, &errstr);
+			if (errstr)
+				errx(1, "ldap port number is %s: %s", errstr, optarg);
+			break;
+#endif
 		case 'm':
 			markall = 1;
 			break;
+		case 'n':
+			recipient_notification = optarg;
+			break;
+		case 'o':
+			notify_only = optarg;
+			break;
+		case 'P':
+			pidfile = optarg;
+			break;
 		case 'p':
 			port = optarg;
 			break;
-		case 'd':
-			tmpdir = optarg;
-			break;
 		case 'T':
 			if (!strcmp(optarg, "g+r+"))
 				tmpfiles_gr = 1;
@@ -335,25 +429,6 @@ main(int argc, char *argv[])
 			else
 				errx(1, "unknown temp file option %s", optarg);
 			break;
-		case 'b':
-			backend = optarg;
-			decode_backend(backend);
-			break;
-		case 'e':
-			logfile = optarg;
-			break;
-		case 'n':
-			recipient_notification = optarg;
-			break;
-		case 'o':
-			notify_only = optarg;
-			break;
-		case 'C':
-			cfgfile = optarg;
-			break;
-		case 'P':
-			pidfile = optarg;
-			break;
 		case 't':
 			chrootdir = optarg;
 			break;
@@ -365,9 +440,6 @@ main(int argc, char *argv[])
 			ldapurl = optarg;
 			break;
 #endif
-		case 'g':
-			group = optarg;
-			break;
 		case 'V':
 			printf("%s %s\n", __progname, VERSION);
 #ifdef ENABLE_LDAP
@@ -380,40 +452,13 @@ main(int argc, char *argv[])
 #endif
 			exit(0);
 			/* NOTREACHED */
-		case 'f':
-			if (getuid()) {
-				getrlimit(RLIMIT_NOFILE, &rl);
-				rlimit_max = rl.rlim_max;
-			} else
-				rlimit_max = RLIM_INFINITY;
-			rlimit_nofile = (int)strtonum(optarg, 1LL, rlimit_max, &errstr);
-			if (errstr)
-				errx(1, "number of files is %s: %s", errstr, optarg);
-			break; 
-		case 'a':
-			spamprefix = optarg;
-			break;
-		case 'k':
-			keep_tmpfiles = 1;
+		case 'v':
+			++verbose;
 			break;
 #ifdef ENABLE_LDAP
-		case 'D':
-			binddn = optarg;
-			break;
-		case 'h':
-			ldaphost = optarg;
-			break;
-		case 'L':
-			ldapport = strtonum(optarg, 1, 65535, &errstr);
-			if (errstr)
-				errx(1, "ldap port number is %s: %s", errstr, optarg);
-			break;
 		case 'w':
 			bindpasswd = optarg;
 			break;
-		case 'B':
-			searchbase = optarg;
-			break;
 #endif
 		case 'x':
 			xmode = 1;
@@ -435,6 +480,36 @@ main(int argc, char *argv[])
 	if (!nbe)
 		errx(1, "no backends defined");
 
+	/* Set default values if some variables are not set */
+
+	if (port == NULL)
+		port = strdup(_PATH_PORT);
+	if (tmpdir == NULL)
+		tmpdir = strdup(_PATH_TMPDIR);
+	if (pidfile == NULL)
+		pidfile = strdup(_PATH_PIDFILE);
+	if (backend == NULL)
+		backend = strdup("clamd");
+	if (virus_strategy == -1)
+		virus_strategy = STRATEGY_DISCARD;
+	if (spam_strategy == -1)
+		spam_strategy = STRATEGY_MARK;
+	if (unwanted_strategy == -1)
+		unwanted_strategy = STRATEGY_MARK;
+	if (error_strategy == -1)
+		error_strategy = STRATEGY_TEMPFAIL;
+	if (log_facility == LOG_UNSET)
+		log_facility = LOG_MAIL;
+
+#ifdef ENABLE_LDAP
+	if (ldaphost == NULL)
+		ldaphost = strdup(DEFLT_LDAPHOST);
+	if (ldapport == -1)
+		ldapport = DEFLT_LDAPPORT;
+	if (ldap_use_tls == -1)
+		ldap_use_tls = 0;
+#endif
+
 	if (xmode) {
 		printf("backends: ");
 		for (n = 0; n < nbe; n++)
@@ -445,13 +520,27 @@ main(int argc, char *argv[])
 		printf("chroot: %s\n", chrootdir != NULL ? chrootdir : "not set");
 		printf("recipient-notification: %s\n", recipient_notification != NULL ? recipient_notification : "not set");
 		printf("spam-subject-prefix: %s\n", spamprefix != NULL ? spamprefix : "not set");
-		printf("port: %s\n", port != NULL ? port : "not set");
-		printf("tmpdir: %s\n", tmpdir != NULL ? tmpdir : "not set");
-		printf("pidfile: %s\n", pidfile != NULL ? pidfile : "not set");
+		printf("port: %s\n", port);
+		printf("tmpdir: %s\n", tmpdir);
+		printf("pidfile: %s\n", pidfile);
 		printf("logfile: %s\n", logfile != NULL ? logfile : "not set");
+		if (rlimit_nofile != 0)
+			printf("maxfiles: %lld\n", (long long)rlimit_nofile);
+		else
+			printf("maxfiles: not set\n"); 
+		if (rlimit_nproc != 0)
+			printf("maxprocs: %lld\n", (long long)rlimit_nproc);
+		else
+			printf("maxprocs: not set\n");
+		if (ex_timeout != -1)
+			printf("timeout: %d\n", ex_timeout);
+		else
+			printf("timeout: not set\n");
 #ifdef ENABLE_LDAP
 		printf("ldapurl: %s\n", ldapurl != NULL ? ldapurl : "not set");
-		printf("ldaphost: %s\n", ldaphost != NULL ? ldaphost : "not set");
+		printf("ldaphost: %s\n", ldaphost);
+		printf("ldapport: %s\n", ldapport);
+		printf("use tls: %s\n", ldap_use_tls ? "yes" : "no");
 		printf("searchbase: %s\n", searchbase != NULL ? searchbase : "not set");
 		printf("binddn: %s\n", binddn != NULL ? binddn : "not set");
 		printf("bindpasswd: %s\n", bindpasswd != NULL ? bindpasswd : "not set");
@@ -507,35 +596,7 @@ main(int argc, char *argv[])
 		return 0;
 	}
 
-	/* Set default values if some variables are not set */
-
-	if (port == NULL)
-		port = _PATH_PORT;
-	if (tmpdir == NULL)
-		tmpdir = _PATH_TMPDIR;
-	if (pidfile == NULL)
-		pidfile = _PATH_PIDFILE;
-	if (backend == NULL)
-		backend = "clamd";
-	if (virus_strategy == -1)
-		virus_strategy = STRATEGY_DISCARD;
-	if (spam_strategy == -1)
-		spam_strategy = STRATEGY_MARK;
-	if (unwanted_strategy == -1)
-		unwanted_strategy = STRATEGY_MARK;
-	if (error_strategy == -1)
-		error_strategy = STRATEGY_TEMPFAIL;
-	if (log_facility == LOG_UNSET)
-		log_facility = LOG_MAIL;
-
 #ifdef ENABLE_LDAP
-	if (ldaphost == NULL)
-		ldaphost = DEFLT_LDAPHOST;
-	if (ldapport == -1)
-		ldapport = DEFLT_LDAPPORT;
-	if (ldap_use_tls == -1)
-		ldap_use_tls = 0;
-
 	if (ldapurl == NULL) {
 		if ((ld = ldap_init(ldaphost, ldapport)) == NULL)
 			errx(1, "No directory server at %s, %m", ldaphost);
@@ -618,6 +679,62 @@ main(int argc, char *argv[])
 		}
 	}
 
+	/* Analye a signe file if requested by the user, else normal operation */
+
+	if (fnam != NULL) {
+		struct be_data *data;
+		char		scanfn[PATH_MAX];
+		char		virus[MAXLEN];
+		char		reason[MAXLEN];
+		double		score, threshold;
+		int		scanresult;
+
+		data = malloc(sizeof(struct be_data) * nbe);
+		for (n = 0; n < nbe; n++) {
+			scanresult = SCAN_OK;
+			if (be[n].type & BE_SCAN_VIRUS) {
+				warnx("scanning %s for viruses using backend %s", fnam, be[n].name);
+				strlcpy(scanfn, fnam, sizeof(scanfn));
+				bzero(virus, sizeof(virus));
+				scanresult = be[n].be_scan.be_scan_virus(&data[n], scanfn, NULL, virus, sizeof(virus));
+
+			} else if (be[n].type & BE_SCAN_SPAM) {
+				warnx("scanning %s for spam using backend %s", fnam, be[n].name);
+				strlcpy(scanfn, fnam, sizeof(scanfn));
+				scanresult = be[n].be_scan.be_scan_spam(&data[n], scanfn, NULL, &score, &threshold);
+
+			} else if (be[n].type & BE_SCAN_UNWANTED) {
+				warnx("scanning %s for unwanted content using backend %s", fnam, be[n].name);
+				strlcpy(scanfn, fnam, sizeof(scanfn));
+				bzero(reason, sizeof(reason));
+				scanresult = be[n].be_scan.be_scan_unwanted(&data[n], scanfn, NULL, reason, sizeof(reason));
+			}
+			switch (scanresult) {
+			case SCAN_OK:
+				warnx("file scanned ok");
+				break;
+			case SCAN_ERROR:
+				warnx("error during scan");
+				break;
+			case SCAN_VIRUS:
+				warnx("virus '%s' detected in file", virus);
+				break;
+			case SCAN_SPAM:
+				warnx("classified as spam, score = %f, threshold %f", score, threshold);
+				break;
+			case SCAN_UNWANTED:
+				warnx("unwanted content, reson = %s", reason);
+				break;
+			}
+		}
+
+		free(data);
+#ifdef ENABLE_LDAP
+		ldap_unbind(ld);
+#endif
+		return 0;
+	}
+
 	uid = getuid();
 	gid = getgid();
 
@@ -642,85 +759,112 @@ main(int argc, char *argv[])
 	if (verbose == 0 &&  daemon(0, 0))
 		err(1, "can't run as daemon");
 
-	/* Setup imsg stuff, this is a one-way com only */
+	while (!quit) {
 
-	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe_m2e) == -1)
-		errx(1, "imsg setup failed");
-	if ((flags = fcntl(pipe_m2e[0], F_GETFL, 0)) == -1)
-		errx(1, "fcntl failed");
-	flags |= O_NONBLOCK;
-	if (fcntl(pipe_m2e[0], F_SETFL, flags) == -1)
-		errx(1, "fcntl can't set flags");
-	if ((flags = fcntl(pipe_m2e[1], F_GETFL, 0)) == -1)
-		errx(1, "fcntl failed");
-	flags |= O_NONBLOCK;
-	if (fcntl(pipe_m2e[1], F_SETFL, flags) == -1)
-		errx(1, "fcntl can't set flags");
+		/* Setup imsg stuff, this is a one-way com only */
 
-	/*
-	 * Fork into two processes, one to run privileged, one to do the
-	 * work.
-	 */
+		if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe_m2e) == -1)
+			errx(1, "imsg setup failed");
+		if ((flags = fcntl(pipe_m2e[0], F_GETFL, 0)) == -1)
+			errx(1, "fcntl failed");
+		flags |= O_NONBLOCK;
+		if (fcntl(pipe_m2e[0], F_SETFL, flags) == -1)
+			errx(1, "fcntl can't set flags");
+		if ((flags = fcntl(pipe_m2e[1], F_GETFL, 0)) == -1)
+			errx(1, "fcntl failed");
+		flags |= O_NONBLOCK;
+		if (fcntl(pipe_m2e[1], F_SETFL, flags) == -1)
+			errx(1, "fcntl can't set flags");
 
-	child_pid = e_main(uid, gid, pipe_m2e);
+		/*
+		 * Fork into two processes, one to run privileged, one to do the
+		 * work.
+		 */
 
-	/* We are the privileged process */
+		if (verbose)
+			warnx("start unprivileged child process");
+
+		child_quit = 0;
+		child_pid = e_main(uid, gid, pipe_m2e);
+
+		/* We are the privileged process */
 
 #ifndef __linux
-	setproctitle("parent");
+		setproctitle("parent");
 #endif
 
-	signal(SIGCHLD, sighdlr);
-	signal(SIGINT, sighdlr);
-	signal(SIGTERM, sighdlr);
+		signal(SIGCHLD, sighdlr);
+		signal(SIGINT, sighdlr);
+		signal(SIGTERM, sighdlr);
+		signal(SIGHUP, sighdlr);
 
-	close(pipe_m2e[1]);
+		close(pipe_m2e[1]);
 
-	if ((ibuf_e = malloc(sizeof(struct imsgbuf))) == NULL)
-		errx(1, "memory error");
+		if ((ibuf_e = malloc(sizeof(struct imsgbuf))) == NULL)
+			errx(1, "memory error");
 
-	imsg_init(ibuf_e, pipe_m2e[0]);
+		imsg_init(ibuf_e, pipe_m2e[0]);
 
-	while (!quit) {
-		pfd[0].fd = ibuf_e->fd;
-		pfd[0].events = POLLIN;
-		if ((nfds = poll(pfd, 1, 1000)) == -1) {
-			if (errno != EINTR) {
-				syslog(LOG_WARNING, "main: poll error");
-				/* quit = 1; */
+		while (!quit && !child_quit) {
+			pfd[0].fd = ibuf_e->fd;
+			pfd[0].events = POLLIN;
+			if ((nfds = poll(pfd, 1, 1000)) == -1) {
+				if (errno != EINTR) {
+					syslog(LOG_WARNING, "main: poll error");
+					/* quit = 1; */
+				}
+			} else if (nfds > 0 && pfd[0].revents & POLLIN) {
+				/* nfds --; */
+				if (dispatch_imsg(ibuf_e, 0) == -1)
+					quit = 1;
 			}
-		} else if (nfds > 0 && pfd[0].revents & POLLIN) {
-			/* nfds --; */
-			if (dispatch_imsg(ibuf_e, 0) == -1)
-				quit = 1;
-		}
 
 #ifdef ENABLE_PF
-		if (pftable_timeout())
-			syslog(LOG_WARNING, "can not timeout pf tables");
+			if (pftable_timeout())
+				syslog(LOG_WARNING, "can not timeout pf tables");
 #endif
+			if (child_quit)
+				warnx("child process terminated");
+		}
+
+		if (quit && !child_quit) {	/* XXX is this necessary? */
+			sleep(1);
+			if (!child_quit) {
+				warnx("signalling child process to terminate");
+				kill(child_pid, SIGTERM);
+			}
+		}
+
+		if (verbose)
+			warnx("waiting for all child processes to terminate");
+
+		do {
+			if ((pid = wait(NULL)) == -1 &&
+			    errno != EINTR && errno != ECHILD)
+				errx(1, "wait");
+		} while (pid != child_pid || (pid == -1 && errno == EINTR));
+
+		if (verbose)
+			warnx("child processes have terminated");
+
+		msgbuf_clear(&ibuf_e->w);
+		free(ibuf_e);
+		close(pipe_m2e[0]);
+
+		if (unlink(pidfile))
+			syslog(LOG_ERR, "can not unlink pidfile %s", pidfile);
+
+		signal(SIGCHLD, SIG_DFL);
+		signal(SIGINT, SIG_DFL);
+		signal(SIGTERM, SIG_DFL);
+		signal(SIGHUP, SIG_DFL);
 	}
 
-	kill(child_pid, SIGTERM);
-
+	closelog();
 #ifdef ENABLE_LDAP
 	ldap_unbind(ld);
 #endif	
 
-	do {
-		if ((pid = wait(NULL)) == -1 &&
-		    errno != EINTR && errno != ECHILD)
-			errx(1, "wait");
-	} while (pid != child_pid || (pid == -1 && errno == EINTR));
-
-	msgbuf_clear(&ibuf_e->w);
-	free(ibuf_e);
-
-	if (!verbose) {
-		if (unlink(pidfile))
-			syslog(LOG_ERR, "can not unlink pidfile %s", pidfile);
-		closelog();
-	}
 	return 0;
 }
 
@@ -748,6 +892,8 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 			break;
 
 		switch (imsg.hdr.type) {
+		case IMSG_RECONFIGURE:
+			break;
 #ifdef ENABLE_PF 
 		case IMSG_PFTABLE_ADD:
 			if (imsg.hdr.len != IMSG_HEADER_SIZE +
